import re
from typing import Optional

from plurk.apis.base import BaseApi
from plurk.exceptions import validate_resp, RespValidationError
from plurk.models import ChannelResp, NoDataChannelResp, UserChannel
from plurk.utils import read_jsonp


def update_new_offset(url: str, new_offset: Optional[int] = None):
    """Update the new_offset query param value in URL and return the new URL
    """
    if new_offset:
        url = re.sub(r'new_offset=-?\d+', f'new_offset={new_offset}', url)
    return url


class Realtime(BaseApi):
    def get_user_channel(self):
        """Returns the user channel info for the curren user.

        You'll get an URL from `/APP/Realtime/getUserChannel` and you do GET requests
        to this URL to get new data.
        Your request will sleep for about 50 seconds before returning a response if there
        is no new data added to your channel. You won't get notifications on responses that
        the logged in user adds, but you will get notifications for new plurks.

        The comet channel will return JSONP format response.
        Comet channel response example: `CometChannel.scriptCallback({"new_offset": 1});`
        """
        endpoint = f'{self.client.base_url}/APP/Realtime/getUserChannel'
        resp = self.client.http_client.get(endpoint)
        validate_resp(resp)
        return UserChannel(**resp.json())

    def get_channel_response(self, comet_channel_url: str, timeout: int = 120, new_offset: Optional[int] = None):
        """Read data from user channel.

        comet_channel_url can be retrieved by calling `get_user_channel()` method
        and access the `comet_channel` attribute.

        By default the request timeout is 120s. The official API doc states the
        request could be pending for about 50s before returning a response.
        """
        endpoint = update_new_offset(comet_channel_url, new_offset)
        resp = self.client.http_client.get(endpoint, timeout=timeout)
        validate_resp(resp)
        jsonp_resp = read_jsonp(resp.text)
        try:
            return ChannelResp(**jsonp_resp)
        except RespValidationError:
            return NoDataChannelResp(**jsonp_resp)



import json
import re

from httpx import HTTPStatusError, Response


class AuthenticationError(Exception):
    """API response with 401 or 403 status code"""


class BadRequestError(Exception):
    """API response with 400 status code"""


class ServerError(Exception):
    """API response with 50x status code"""


class RespValidationError(Exception):
    """API Response is not in the expected format
    """


class UnknownError(Exception):
    """API response with another non-succesfful status code"""


def extract_error_message(r: Response):
    """Normally the API returns a dict with 'error_text' key
    for detailed error message. However, under some circumstances
    (ex. the server fails to parse the request body)
    """
    try:
        msg = r.json()['error_text']
    except KeyError:
        # Response is JSON, but error_text key missing
        msg = r.text
    except json.decoder.JSONDecodeError:
        # Response is not JSON. Usually it's HTML body
        msg_match = re.search('<\\s*p[^>]*>(.*?)<\\s*/\\s*p>', r.text)  # look for <p>msg</p>
        if msg_match:
            msg = msg_match.group(0)
        else:
            msg = ''
    except Exception:
        msg = r.text
    return msg


def validate_resp(r: Response):
    try:
        r.raise_for_status()
    except HTTPStatusError as exc:
        exception_class = None
        extra_msg = ''
        if r.status_code in [401, 403]:
            exception_class = AuthenticationError
        elif r.status_code == 400:
            exception_class = BadRequestError
        elif r.status_code >= 500:
            exception_class = ServerError
        else:
            exception_class = UnknownError
        extra_msg = extract_error_message(r)
        raise exception_class(extra_msg) from exc
    return r



from typing import Dict

from pytest_httpx import HTTPXMock

from plurk import Client

FAKE_APP_CREDENTIALS = ('fake_app_key', 'fake_app_secret')


def test_get(httpx_mock: HTTPXMock, get_emoticons_fixture):
    resp_json = get_emoticons_fixture
    httpx_mock.add_response(
        status_code=200,
        json=resp_json,
    )
    with Client(*FAKE_APP_CREDENTIALS) as client:
        resp = client.emoticons.get()
    assert isinstance(resp, Dict)
    assert get_emoticons_fixture == resp



from .alert import (FriendshipAcceptedAlert, FriendshipPendingAlert,
                    FriendshipRequestAlert, MentionedAlert, MyRespondedAlert,
                    NewFanAlert, NewFriendAlert, PlurkLikedAlert,
                    PlurkReplurkedAlert, PrivatePlurkAlert)
from .blocks import GetBlocksResp
from .general_resps import ActionResp
from .karma_stats import KarmaStats
from .plurk import NewPlurk, Plurk
from .profile import OwnProfile, PublicProfile
from .response import NewResponse, Response, ResponsesGetResp
from .search import PlurkSearchResp, UserSearchResp
from .timeline import GetPlurkResp, GetPlurksResp, UploadPictureResp
from .user_channel import ChannelResp, NoDataChannelResp, UserChannel
from .user_data import (PublicUserData, SimpleUserData, UpdateAvatarResp,
                        UserData)

__all__ = [
    'FriendshipAcceptedAlert',
    'FriendshipPendingAlert',
    'FriendshipRequestAlert',
    'MentionedAlert',
    'MyRespondedAlert',
    'NewFanAlert',
    'NewFriendAlert',
    'PlurkLikedAlert',
    'PlurkReplurkedAlert',
    'PrivatePlurkAlert',
    'GetBlocksResp',
    'ChannelResp',
    'NoDataChannelResp',
    'KarmaStats',
    'NewPlurk',
    'NewResponse',
    'Response',
    'ResponsesGetResp',
    'OwnProfile',
    'Plurk',
    'PublicProfile',
    'PublicUserData',
    'PlurkSearchResp',
    'UserSearchResp',
    'GetPlurkResp',
    'GetPlurksResp',
    'SimpleUserData',
    'ActionResp',
    'UserChannel',
    'UserData',
    'UpdateAvatarResp',
    'UploadPictureResp',
]



from enum import Enum


class Filter(str, Enum):
    """Filter is a optional parameter for some endpoints that returns multiple plurks.
    Possible values:
    - `my`
    - `responded`
    - `private`
    - `favorite`
    - `replurked`
    - `mentioned` (only available to premium users)
    """
    MY = 'my'
    RESPONSED = 'responded'
    PRIVATE = 'private'
    FAVORITE = 'favorite'
    REPLURKED = 'replurked'
    MENTIONED = 'mentioned'



from .client import Client
from .enums import (AbuseCategory, AvatarSize, ChannelDataEntryType, Language,
                    Privacy, Qualifier, Relationship)


__all__ = [
    'AbuseCategory',
    'AvatarSize',
    'ChannelDataEntryType',
    'Client',
    'Language',
    'Privacy',
    'Qualifier',
    'Relationship',
]



from datetime import datetime
from typing import Dict, List, Literal, Optional

from plurk.enums import ChannelDataEntryType, Language, Qualifier
from plurk.models.base import RespBase
from plurk.models.plurk import Plurk
from plurk.models.user_data import PublicUserData, SimpleUserData
from plurk.utils import parse_time_validator


class NewResponse(RespBase):
    """Part of data returned from comet channel URL
    """
    plurk_id: int
    plurk: Plurk
    user: Dict[str, SimpleUserData]
    type: Literal[ChannelDataEntryType.NEW_RESPONSE]


class Response(RespBase):
    """Data of a plurk response
    """
    coins: Optional[int]
    """How many plurk coins are gifted to the response.
    """
    content: str
    content_raw: str
    editability: int
    """Whether the response can be edited by the current user. The value is either 0 or 1.
    `0`: It can not be edited.
    `1`: It can be edited.
    """
    id: int
    """The response id.
    """
    lang: Language
    last_edited: Optional[datetime]
    plurk_id: int
    posted: datetime
    qualifier: Qualifier
    qualifier_translated: str
    user_id: int
    with_random_emos: Optional[bool]

    _parse_last_edited = parse_time_validator('last_edited')
    _parse_posted = parse_time_validator('posted')


class ResponsesGetResp(RespBase):
    """Resp of `/APP/Responses/get`
    """
    friends: Dict[str, PublicUserData]
    response_count: int
    responses_seen: int
    responses: List[Response]



from typing import List

from plurk.models.base import RespBase
from plurk.models.user_data import SimpleUserData


class GetBlocksResp(RespBase):
    """Resp of `/APP/Blocks/get`
    """
    total: int
    users: List[SimpleUserData]



from typing import List

from pytest_httpx import HTTPXMock

from plurk import Client
from plurk.models import ActionResp, PublicUserData

FAKE_APP_CREDENTIALS = ('fake_app_key', 'fake_app_secret')
FAKE_CLIQUE_NAME = 'fake_clique_name'
FAKE_NEW_NAME = 'fake_new_name'
FAKE_USER_ID = -1


def test_get_cliques(httpx_mock: HTTPXMock, get_cliques_resp_fixture):
    resp_json = get_cliques_resp_fixture
    httpx_mock.add_response(
        status_code=200,
        json=resp_json,
    )
    with Client(*FAKE_APP_CREDENTIALS) as client:
        get_cliques_resp = client.cliques.get_cliques()
    assert isinstance(get_cliques_resp, list)
    assert get_cliques_resp_fixture == get_cliques_resp


def test_get_clique(httpx_mock: HTTPXMock, public_user_data_list_fixture: List[PublicUserData]):
    resp_json = [i.dict_original() for i in public_user_data_list_fixture]
    httpx_mock.add_response(
        status_code=200,
        json=resp_json,
    )
    with Client(*FAKE_APP_CREDENTIALS) as client:
        resp = client.cliques.get_clique(FAKE_CLIQUE_NAME)
    assert isinstance(resp, list)
    assert [i.dict_original() for i in resp] == resp_json


def test_create_clique(httpx_mock: HTTPXMock, action_resp_fixture):
    resp_json = action_resp_fixture.dict_original()
    httpx_mock.add_response(
        status_code=200,
        json=resp_json,
    )
    with Client(*FAKE_APP_CREDENTIALS) as client:
        resp = client.cliques.create_clique(FAKE_CLIQUE_NAME)
    assert isinstance(resp, ActionResp)
    assert action_resp_fixture.dict_original() == resp_json


def test_rename_clique(httpx_mock: HTTPXMock, action_resp_fixture):
    resp_json = action_resp_fixture.dict_original()
    httpx_mock.add_response(
        status_code=200,
        json=resp_json,
    )
    with Client(*FAKE_APP_CREDENTIALS) as client:
        resp = client.cliques.rename_clique(FAKE_CLIQUE_NAME, FAKE_NEW_NAME)
    assert isinstance(resp, ActionResp)
    assert action_resp_fixture.dict_original() == resp_json


def test_add(httpx_mock: HTTPXMock, action_resp_fixture):
    resp_json = action_resp_fixture.dict_original()
    httpx_mock.add_response(
        status_code=200,
        json=resp_json,
    )
    with Client(*FAKE_APP_CREDENTIALS) as client:
        resp = client.cliques.add(FAKE_CLIQUE_NAME, FAKE_USER_ID)
    assert isinstance(resp, ActionResp)
    assert action_resp_fixture.dict_original() == resp_json


def test_remove(httpx_mock: HTTPXMock, action_resp_fixture):
    resp_json = action_resp_fixture.dict_original()
    httpx_mock.add_response(
        status_code=200,
        json=resp_json,
    )
    with Client(*FAKE_APP_CREDENTIALS) as client:
        resp = client.cliques.remove(FAKE_CLIQUE_NAME, FAKE_USER_ID)
    assert isinstance(resp, ActionResp)
    assert action_resp_fixture.dict_original() == resp_json



from enum import Enum


class Relationship(str, Enum):
    NOT_SAYING = 'not_saying'
    SINGLE = 'single'
    MARRIED = 'married'
    DIVORCED = 'divorced'
    ENGAGED = 'engaged'
    IN_RELATIONSHIP = 'in_relationship'
    COMPLICATED = 'complicated'
    WIDOWED = 'widowed'
    UNSTABLE_RELATIONSHIP = 'unstable_relationship'
    OPEN_RELATIONSHIP = 'open_relationship'



from typing import Optional

from plurk.apis.base import BaseApi
from plurk.enums import Qualifier
from plurk.exceptions import validate_resp
from plurk.models import ActionResp, Response, ResponsesGetResp
from plurk.utils import build_params


class Responses(BaseApi):
    def get(
        self,
        plurk_id: int,
        from_response: Optional[int] = None,
        minimal_data: Optional[bool] = None,
        count: Optional[int] = None,
    ):
        """Fetches responses for plurk with `plurk_id` and some basic info about the users.
        """
        endpoint = f'{self.client.base_url}/APP/Responses/get'
        params = build_params(
            plurk_id=plurk_id,
            from_response=from_response,
            minimal_data=minimal_data,
            count=count,
        )
        resp = self.client.http_client.get(endpoint, params=params)
        validate_resp(resp)
        return ResponsesGetResp(**resp.json())

    def response_add(self, plurk_id: int, content: str, qualifier: Qualifier = Qualifier.EMPTY):
        """Adds a responses to `plurk_id`. Language is inherited from the plurk.
        """
        endpoint = f'{self.client.base_url}/APP/Responses/responseAdd'
        payload = build_params(
            plurk_id=plurk_id,
            content=content,
            qualifier=qualifier,
        )
        resp = self.client.http_client.post(endpoint, json=payload)
        validate_resp(resp)
        return Response(**resp.json())

    def response_delete(self, plurk_id: int, response_id: int):
        """Deletes a response. A user can delete own responses or responses that are posted to own plurks.
        """
        endpoint = f'{self.client.base_url}/APP/Responses/responseDelete'
        payload = build_params(
            response_id=response_id,
            plurk_id=plurk_id,
        )
        resp = self.client.http_client.post(endpoint, json=payload)
        validate_resp(resp)
        return ActionResp(**resp.json())



from datetime import datetime
from typing import Any, Dict, List, Union

from pydantic import BaseModel, ValidationError

from plurk.exceptions import RespValidationError
from plurk.utils import format_as_plurk_time


def format_datetime_deep(obj: Union[Dict, List, int, str, bool, datetime, None]):
    if isinstance(obj, datetime):
        return format_as_plurk_time(obj)
    if isinstance(obj, list):
        return [format_datetime_deep(item) for item in obj]
    if isinstance(obj, dict):
        return {k: format_datetime_deep(v) for k, v in obj.items()}
    return obj


class RespBase(BaseModel):
    def __init__(self, *args, **kwargs):
        try:
            super().__init__(*args, **kwargs)
        except ValidationError as exc:
            raise RespValidationError(self.__class__.__name__) from exc

    def dict_original(self) -> Dict[str, Any]:
        """Return a dict representation of the UserData while restoring the original
        data like Plurk API date string format.
        """
        return format_datetime_deep(self.dict())  # type: ignore



from typing import List, Union
from typing_extensions import Annotated

from pydantic import Field

from plurk.models.base import RespBase
from plurk.models.plurk import NewPlurk
from plurk.models.response import NewResponse


ChannelDataEntry = Annotated[Union[NewPlurk, NewResponse], Field(discriminator='type')]


class BaseChannelResp(RespBase):
    """Resp returned from comet channel URL
    """
    new_offset: int


class NoDataChannelResp(BaseChannelResp):
    """Resp returned from comet channel URL that indicates no updates
    """


class ChannelResp(BaseChannelResp):
    """Resp returned from comet channel URL that contains at least one data entry
    """
    data: List[ChannelDataEntry]


class UserChannel(RespBase):
    """Data returned by `APP/Realtime/getUserChannel` endpoint
    """
    channel_name: str
    comet_server: str



from plurk.apis.base import BaseApi
from plurk.exceptions import validate_resp
from plurk.models import ActionResp, GetBlocksResp
from plurk.utils import build_params


class Blocks(BaseApi):
    def get(self):
        r"""Get users that are blocked by the current user.
        """
        endpoint = f'{self.client.base_url}/APP/Blocks/get'
        resp = self.client.http_client.get(endpoint)
        validate_resp(resp)
        return GetBlocksResp(**resp.json())

    def block(self, user_id: int):
        r"""Block a user.
        """
        endpoint = f'{self.client.base_url}/APP/Blocks/block'
        payload = build_params(user_id=user_id)
        resp = self.client.http_client.post(endpoint, json=payload)
        validate_resp(resp)
        return ActionResp(**resp.json())

    def unblock(self, user_id: int):
        r"""Unblock a user.
        """
        endpoint = f'{self.client.base_url}/APP/Blocks/unblock'
        payload = build_params(user_id=user_id)
        resp = self.client.http_client.post(endpoint, json=payload)
        validate_resp(resp)
        return ActionResp(**resp.json())




from enum import Enum
import json
import re
from datetime import datetime
from typing import Optional

from pydantic import validator

PLURK_DATETIME_FORMAT = '%a, %d %b %Y %H:%M:%S GMT'


def build_params(**kwargs):
    """Return a dict built from non-null kwargs.

    If a kwarg is an Enum, the value of enum will be used for building the params.
    Used for building query params or request body.
    """
    ret = {}
    for k, v in kwargs.items():
        if v is None:
            continue
        if isinstance(v, Enum):
            ret[k] = v.value
        else:
            ret[k] = v
    return ret


def format_as_plurk_time(dt: datetime):
    """Return a date string in the format the Plurk API's responses use.
    """
    return dt.strftime(PLURK_DATETIME_FORMAT)


def parse_plurk_time(date_string: Optional[str]):
    """Parse the date string used in Plurk API's responses into datetime object.

    The utility function exists because Plurk API does not use ISO8601 datetime format.
    """
    if not date_string:
        return None
    return datetime.strptime(date_string, PLURK_DATETIME_FORMAT)


def parse_time_validator(field: str):
    """Used for Pydantic models to parse datetime fields before the other validators.

    Setting up the validators to avoid 'invalid datetime format' when initializing a
    model with datetime fields from Plurk API's response data.

    Usage example:

    class UserData(BaseModel):

        join_date: datetime

        _parse_join_date = parse_time_validator('join_date')
    """
    return validator(field, pre=True, allow_reuse=True)(parse_plurk_time)


def read_jsonp(jsonp_str: str):
    json_str_match = re.match('.*?({.*}).*', jsonp_str)
    if json_str_match:
        try:
            return json.loads(json_str_match.group(1))
        except json.decoder.JSONDecodeError:
            pass
    raise ValueError('Input is not a valid JSONP string')



from enum import Enum


class Language(str, Enum):
    """ A plurk's language. Can be the following:
        'en': 'English'
        'tr_ch': '中文 (台灣)'
        'tr_hk': '中文 (香港)'
        'cn': '中文 (中国)'
        'ja': '日本語'
        'ca': 'Català'
        'el': 'Ελληνικά'
        'dk': 'Dansk'
        'de': 'Deutsch'
        'es': 'Español'
        'sv': 'Svenska'
        'nb': 'Norsk bokmål'
        'hi': 'Hindi'
        'ro': 'Română'
        'hr': 'Hrvatski'
        'fr': 'Français'
        'ru': 'Pусский'
        'it': 'Italiano '
        'he': 'עברית'
        'hu': 'Magyar'
        'ne': 'Nederlands'
        'th': 'ไทย'
        'ta_fp': 'Filipino'
        'in': 'Bahasa Indonesia'
        'pl': 'Polski'
        'ar': 'العربية'
        'fi': 'Finnish'
        'tr': 'Türkçe'
        'ga': 'Gaeilge'
        'sk': 'Slovenský'
        'uk': 'українська'
        'fa': 'فارسی
        'pt_BR': 'Português'
    """
    EN = 'en'
    TR_CH = 'tr_ch'
    TR_HK = 'tr_hk'
    CN = 'cn'
    JA = 'ja'
    CA = 'ca'
    EL = 'el'
    DK = 'dk'
    DE = 'de'
    ES = 'es'
    SV = 'sv'
    NB = 'nb'
    HI = 'hi'
    RO = 'ro'
    HR = 'hr'
    FR = 'fr'
    RU = 'ru'
    IT = 'it'
    HE = 'he'
    HU = 'hu'
    NE = 'ne'
    TH = 'th'
    TA_FP = 'ta_fp'
    IN = 'in'
    PL = 'pl'
    AR = 'ar'
    FI = 'fi'
    TR = 'tr'
    GA = 'ga'
    SK = 'sk'
    UK = 'uk'
    FA = 'fa'
    PT_BR = 'pt_BR'



from pytest_httpx import HTTPXMock

from plurk import Client
from plurk.models import ActionResp, GetBlocksResp

FAKE_APP_CREDENTIALS = ('fake_app_key', 'fake_app_secret')
FAKE_USER_ID = -1


def test_get(httpx_mock: HTTPXMock, get_blocks_resp_fixture):
    resp_json = get_blocks_resp_fixture.dict_original()
    httpx_mock.add_response(
        status_code=200,
        json=resp_json,
    )
    with Client(*FAKE_APP_CREDENTIALS) as client:
        get_blocks_resp = client.blocks.get()
    assert isinstance(get_blocks_resp, GetBlocksResp)
    assert get_blocks_resp_fixture.dict_original() == resp_json


def test_block(httpx_mock: HTTPXMock, action_resp_fixture):
    resp_json = action_resp_fixture.dict_original()
    httpx_mock.add_response(
        status_code=200,
        json=resp_json,
    )
    with Client(*FAKE_APP_CREDENTIALS) as client:
        resp = client.blocks.block(FAKE_USER_ID)
    assert isinstance(resp, ActionResp)
    assert action_resp_fixture.dict_original() == resp_json


def test_unblock(httpx_mock: HTTPXMock, action_resp_fixture):
    resp_json = action_resp_fixture.dict_original()
    httpx_mock.add_response(
        status_code=200,
        json=resp_json,
    )
    with Client(*FAKE_APP_CREDENTIALS) as client:
        resp = client.blocks.unblock(FAKE_USER_ID)
    assert isinstance(resp, ActionResp)
    assert action_resp_fixture.dict_original() == resp_json



from typing import Optional

from plurk.apis.base import BaseApi
from plurk.exceptions import validate_resp
from plurk.models import PlurkSearchResp
from plurk.utils import build_params


class PlurkSearch(BaseApi):
    def search(self, query: str, offset: Optional[int] = None):
        """Returns the latest 20 plurks on a search term.
        """
        endpoint = f'{self.client.base_url}/APP/PlurkSearch/search'
        params = build_params(query=query, offset=offset)
        resp = self.client.http_client.get(endpoint, params=params)
        validate_resp(resp)
        return PlurkSearchResp(**resp.json())



from pytest_httpx import HTTPXMock

from plurk import Client
from plurk.models import OwnProfile, PublicProfile


FAKE_APP_CREDENTIALS = ('fake_app_key', 'fake_app_secret')


def test_get_own_profile(httpx_mock: HTTPXMock, own_profile_fixture):
    resp_json = own_profile_fixture.dict_original()
    httpx_mock.add_response(
        status_code=200,
        json=resp_json,
    )
    with Client(*FAKE_APP_CREDENTIALS) as client:
        own_profile = client.profile.get_own_profile()
    assert isinstance(own_profile, OwnProfile)
    assert own_profile.dict_original() == resp_json


def test_get_public_profile(httpx_mock: HTTPXMock, public_profile_fixture):
    resp_json = public_profile_fixture.dict_original()
    httpx_mock.add_response(
        status_code=200,
        json=resp_json,
    )
    with Client(*FAKE_APP_CREDENTIALS) as client:
        public_profile = client.profile.get_public_profile()
    assert isinstance(public_profile, PublicProfile)
    assert public_profile.dict_original() == resp_json



from enum import Enum


class Qualifier(str, Enum):
    """Every plurk has a qualifier and it will be displayed on Plurk site.

    The display for EMPTY, COLON and FREESTYLE is the same on Plurk site, showing a `:` symbol.
    """
    EMPTY = ''
    COLON = ':'
    FREESTYLE = 'freestyle'
    PLAYS = 'plays'
    BUYS = 'buys'
    SELLS = 'sells'
    LOVES = 'loves'
    LIKES = 'likes'
    SHARES = 'shares'
    HATES = 'hates'
    WANTS = 'wants'
    WISHES = 'wishes'
    NEEDS = 'needs'
    HAS = 'has'
    WILL = 'will'
    HOPES = 'hopes'
    ASKS = 'asks'
    WONDERS = 'wonders'
    FEELS = 'feels'
    THINKS = 'thinks'
    DRAWS = 'draws'
    IS = 'is'
    SAYS = 'says'
    EATS = 'eats'
    WRITES = 'writes'
    WHISPERS = 'whispers'



import pytest
from authlib.integrations.base_client.errors import OAuthError
from authlib.integrations.httpx_client import OAuth1Client

from plurk import Client


FAKE_APP_CREDENTIALS = ('fake_app_key', 'fake_app_secret')
FAKE_AUTH_CODE = 'auth_code'
FAKE_AUTH_URL = 'http://auth_url'
FAKE_CLIENT_CREDENTIALS = ('fake_oauth_token', 'fake_oauth_secret')
FAKE_REQUEST_TOKEN = {'oauth_token': '', 'oauth_token_secret': ''}


def test_get_request_token(mocker):
    mocker.patch(
        'authlib.integrations.httpx_client.OAuth1Client.fetch_request_token',
        return_value=FAKE_REQUEST_TOKEN
    )
    with Client(*FAKE_APP_CREDENTIALS) as client:
        request_token = client.get_request_token()
    assert request_token == FAKE_REQUEST_TOKEN


def test_get_request_token_fail(mocker):
    exc_msg = 'fetch_token_denied: Token request failed with code 401'
    mocker.patch(
        'authlib.integrations.httpx_client.OAuth1Client.fetch_request_token',
        side_effect=OAuthError(exc_msg)
    )
    with pytest.raises(OAuthError) as exc:
        with Client(*FAKE_APP_CREDENTIALS) as client:
            client.get_request_token()
    assert exc_msg in str(exc)


def test_get_auth_url(mocker):
    mocker.patch(
        'authlib.integrations.httpx_client.OAuth1Client.create_authorization_url',
        return_value=FAKE_AUTH_URL
    )
    with Client(*FAKE_APP_CREDENTIALS) as client:
        auth_url = client.get_auth_url(FAKE_REQUEST_TOKEN)
    assert auth_url == FAKE_AUTH_URL


def test_get_auth_url_request_token_key_error(mocker):
    with pytest.raises(KeyError):
        with Client(*FAKE_APP_CREDENTIALS) as client:
            client.get_auth_url({})


def test_fetch_access_token(mocker):
    access_token_fixture = {
        'oauth_token': 'fake_oauth_token',
        'oauth_token_secret': 'fake_oauth_token_secret',
    }
    mocker.patch(
        'authlib.integrations.httpx_client.OAuth1Client.fetch_access_token',
        return_value=access_token_fixture,
    )
    with Client(*FAKE_APP_CREDENTIALS) as client:
        access_token = client.fetch_access_token(
            request_token=FAKE_REQUEST_TOKEN,
            oauth_verifier=FAKE_AUTH_CODE,
        )
        assert access_token == access_token_fixture
        assert client.token == access_token_fixture['oauth_token']
        assert client.token_secret == access_token_fixture['oauth_token_secret']

        # Test that the underlying client is properly configured for auth
        assert client.http_client.auth.__dict__ == OAuth1Client(
            *FAKE_APP_CREDENTIALS,
            token=access_token_fixture['oauth_token'],
            token_secret=access_token_fixture['oauth_token_secret'],
        ).auth.__dict__


def test_fetch_access_token_fail(mocker):
    exc_msg = 'fetch_token_denied: Token request failed with code 400'
    mocker.patch(
        'authlib.integrations.httpx_client.OAuth1Client.fetch_access_token',
        side_effect=OAuthError(exc_msg)
    )
    with pytest.raises(OAuthError) as exc:
        with Client(*FAKE_APP_CREDENTIALS) as client:
            client.fetch_access_token(
                request_token=FAKE_REQUEST_TOKEN,
                oauth_verifier=FAKE_AUTH_CODE,
            )
    assert exc_msg in str(exc)



from abc import abstractmethod
from typing import Dict, TypeVar

from authlib.integrations.httpx_client import OAuth1Client

from plurk import apis, oauth
from plurk.exceptions import validate_resp

T = TypeVar('T')


class BaseClient():
    """Base class for clients
    """
    @property
    @abstractmethod
    def http_client_class(self):
        pass

    def __init__(self, app_key: str, app_secret: str, base_url='https://www.plurk.com'):
        self.http_client = None
        self.app_key = app_key
        self.app_secret = app_secret
        self.base_url = base_url
        self.token = None
        self.token_secret = None


class Client(BaseClient):
    """Synchronous client for Plurk API
    """
    @property
    def http_client_class(self):
        return OAuth1Client

    def __enter__(self):
        self.setup_client()
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.http_client.close()

    def setup_client(self):
        if self.http_client:
            self.http_client.close()
        self.http_client = oauth.get_oauth_client(
            self.app_key,
            self.app_secret,
            self.token,
            self.token_secret,
        )

    def get_request_token(self):
        return oauth.get_request_token(
            self.http_client,
            request_token_url=f'{self.base_url}/OAuth/request_token',
        )

    def get_auth_url(self, request_token: Dict):
        return oauth.get_auth_url(
            self.http_client,
            authenticate_url=f'{self.base_url}/OAuth/authorize',
            request_token=request_token
        )

    def fetch_access_token(self, request_token: Dict, oauth_verifier: str):
        access_token = oauth.fetch_access_token(
            self.app_key, self.app_secret,
            access_token_url=f'{self.base_url}/OAuth/access_token',
            request_token=request_token,
            oauth_verifier=oauth_verifier,
        )
        self.token = access_token['oauth_token']
        self.token_secret = access_token['oauth_token_secret']
        self.setup_client()
        return access_token

    def set_access_token(self, token: str, token_secret: str):
        self.token = token
        self.token_secret = token_secret
        self.setup_client()

    def checkToken(self):
        endpoint = f'{self.base_url}/APP/checkToken'
        resp = self.http_client.post(endpoint)
        validate_resp(resp)
        return resp.json()

    def expireToken(self):
        endpoint = f'{self.base_url}/APP/expireToken'
        resp = self.http_client.post(endpoint)
        validate_resp(resp)
        return resp.json()

    def checkTime(self):
        endpoint = f'{self.base_url}/APP/checkTime'
        resp = self.http_client.post(endpoint)
        validate_resp(resp)
        return resp.json()

    def echo(self, data: Dict[str, T]) -> Dict[str, T]:
        endpoint = f'{self.base_url}/APP/echo'
        resp = self.http_client.post(endpoint, data=data)
        validate_resp(resp)
        return resp.json()

    @property
    def users(self):
        return apis.Users(self)

    @property
    def profile(self):
        return apis.Profile(self)

    @property
    def realtime(self):
        return apis.Realtime(self)

    @property
    def timeline(self):
        return apis.Timeline(self)

    @property
    def responses(self):
        return apis.Responses(self)

    @property
    def friends_fans(self):
        return apis.FriendsFans(self)

    @property
    def alerts(self):
        return apis.Alerts(self)

    @property
    def plurk_search(self):
        return apis.PlurkSearch(self)

    @property
    def user_search(self):
        return apis.UserSearch(self)

    @property
    def emoticons(self):
        return apis.Emoticons(self)

    @property
    def blocks(self):
        return apis.Blocks(self)

    @property
    def cliques(self):
        return apis.Cliques(self)

    @property
    def helpers(self):
        return apis.Helper(self)



from typing import List

from pydantic import parse_obj_as

from plurk.apis.base import BaseApi
from plurk.exceptions import validate_resp
from plurk.models import ActionResp, PublicUserData
from plurk.utils import build_params


class Cliques(BaseApi):
    def get_cliques(self) -> List[str]:
        r"""Return a list of clique names.
        """
        endpoint = f'{self.client.base_url}/APP/Cliques/getCliques'
        resp = self.client.http_client.get(endpoint)
        validate_resp(resp)
        return resp.json()

    def get_clique(self, clique_name):
        r"""Get users in a clique.
        """
        endpoint = f'{self.client.base_url}/APP/Cliques/getClique'
        params = build_params(clique_name=clique_name)
        resp = self.client.http_client.get(endpoint, params=params)
        validate_resp(resp)
        return parse_obj_as(List[PublicUserData], resp.json())

    def create_clique(self, clique_name: str):
        r"""Create a new clique.
        """
        endpoint = f'{self.client.base_url}/APP/Cliques/createClique'
        payload = build_params(clique_name=clique_name)
        resp = self.client.http_client.post(endpoint, json=payload)
        validate_resp(resp)
        return ActionResp(**resp.json())

    def rename_clique(self, clique_name: str, new_name: str):
        r"""Rename a clique.
        """
        endpoint = f'{self.client.base_url}/APP/Cliques/renameClique'
        payload = build_params(clique_name=clique_name, new_name=new_name)
        resp = self.client.http_client.post(endpoint, json=payload)
        validate_resp(resp)
        return ActionResp(**resp.json())

    def add(self, clique_name: str, user_id: int):
        r"""Add a user to the clique.
        """
        endpoint = f'{self.client.base_url}/APP/Cliques/add'
        payload = build_params(clique_name=clique_name, user_id=user_id)
        resp = self.client.http_client.post(endpoint, json=payload)
        validate_resp(resp)
        return ActionResp(**resp.json())

    def remove(self, clique_name: str, user_id: int):
        r"""Add a user to the clique.
        """
        endpoint = f'{self.client.base_url}/APP/Cliques/remove'
        payload = build_params(clique_name=clique_name, user_id=user_id)
        resp = self.client.http_client.post(endpoint, json=payload)
        validate_resp(resp)
        return ActionResp(**resp.json())



from .abuse_category import AbuseCategory
from .alert_type import AlertType
from .avatar_size import AvatarSize
from .channel_data_entry_type import ChannelDataEntryType
from .filter import Filter
from .language import Language
from .privacy import Privacy
from .qualifier import Qualifier
from .relationship import Relationship

__all__ = [
    'AbuseCategory',
    'AlertType',
    'AvatarSize',
    'ChannelDataEntryType',
    'Filter',
    'Language',
    'Qualifier',
    'Privacy',
    'Relationship',
]



from plurk.apis.base import BaseApi
from plurk.exceptions import validate_resp
from plurk.models.profile import OwnProfile, PublicProfile


class Profile(BaseApi):
    def get_own_profile(self):
        """Returns data that's private for the current user.

        This can be used to construct a profile and render a timeline of the latest plurks.
        """
        endpoint = f'{self.client.base_url}/APP/Profile/getOwnProfile'
        resp = self.client.http_client.get(endpoint)
        validate_resp(resp)
        return OwnProfile(**resp.json())

    def get_public_profile(self):
        """Fetches public information such as a user's public plurks and basic information.

        Fetches also if the current user is following the user, are friends with or is a fan.

        WARNING: The API is currently unavailable as the endpoint returns 404 status code.
        """
        endpoint = f'{self.client.base_url}/APP/Users/getPublicProfile'
        resp = self.client.http_client.get(endpoint)
        validate_resp(resp)
        return PublicProfile(**resp.json())



from datetime import datetime
from typing import List

from plurk.models.base import RespBase


class KarmaStats(RespBase):
    """Data returned by `/APP/Users/getKarmaStats` endpoint
    """
    current_karma: float
    karma_trend: List[str]
    karma_graph: str
    karma_fall_reason: str

    def karma_trend_to_tuples(self):
        """The raw karma_trend data is formatted as '{unixtimestamp}-{karma_value}'.
        The helper function returns a list of tuples based on raw data as
        datetime and karma-value tuples.
        """
        return [
            (datetime.fromtimestamp(int(i.split('-')[0])), float(i.split('-')[1]))
            for i in self.karma_trend
        ]



from datetime import datetime
from typing import BinaryIO, List, Optional, Union

from plurk.apis.base import BaseApi
from plurk.enums import AbuseCategory, Filter, Language, Qualifier
from plurk.exceptions import validate_resp
from plurk.models import (ActionResp, GetPlurkResp, GetPlurksResp, Plurk,
                          UploadPictureResp)
from plurk.utils import build_params


class Timeline(BaseApi):
    def get_plurk(self, plurk_id: int, favorers_detail=False, limited_detail=False, replurkers_detail=False):
        """Return a plurk by its plurk_id.

        Note: the response schema described in offical API doc is incorrect. See `GetPlurkResp` model
        for the actual data schema.

        Args:
            plurk_id: the unique id of a plurk.
            favorers_detail: if true, detailed users information about all favorers of this plurk will be
            included into "plurk_users".
            limited_tail: if true, detailed users information about recepients of this plurk will be included
            into "plurk_users" (if this plurk is private).
            replurkers_detail: if true, detailed users information about all replurkers of this plurk will be
            included into "plurk_users".
        """
        endpoint = f'{self.client.base_url}/APP/Timeline/getPlurk'
        params = build_params(
            plurk_id=plurk_id,
            favorers_detail=favorers_detail,
            limited_detail=limited_detail,
            replurkers_detail=replurkers_detail,
        )
        resp = self.client.http_client.get(endpoint, params=params)
        validate_resp(resp)
        return GetPlurkResp(**resp.json())

    def get_plurks(
                self,
                offset: Optional[Union[str, datetime]] = None,
                limit: int = 20,
                filter: Optional[Filter] = None,  # pylint: disable=redefined-builtin
                favorers_detail=False,
                limited_detail=False,
                replurkers_detail=False,
            ):
        """Return plurks on timeline.

        Args:
            offset: return plurks older than offset, formatted as `YYYY-mm-DDTHH:MM:SS`. The function can also
                accept a datetime object and convert it automatically to the format for the API.
            limit: how many plurks should be returned. Default: `20`.
            filter: can be `my`, `responded`, `private`, `favorite`, `replurked` or `mentioned`; `mentioned` is only
                available for users with `user.premium = true`.
            favorers_detail: If true, detailed users information about all favorers of this plurk will be included into
                "plurk_users".
            limited_tail: If true, detailed users information about recepients of this plurk will be included into
                "plurk_users" (if this plurk is private).
            replurkers_detail: If true, detailed users information about all replurkers of this plurk will be included
                into "plurk_users".
        """
        endpoint = f'{self.client.base_url}/APP/Timeline/getPlurks'
        if isinstance(offset, datetime):
            offset = offset.strftime('%Y-%m-%dT%H:%M:%S')
        params = build_params(
            offset=offset,
            limit=limit,
            filter=filter,
            favorers_detail=favorers_detail,
            limited_detail=limited_detail,
            replurkers_detail=replurkers_detail,
        )
        resp = self.client.http_client.get(endpoint, params=params)
        validate_resp(resp)
        return GetPlurksResp(**resp.json())

    def get_unread_plurks(
                self,
                offset: Optional[Union[str, datetime]] = None,
                limit: int = 20,
                filter: Optional[Filter] = None,  # pylint: disable=redefined-builtin
                favorers_detail=False,
                limited_detail=False,
                replurkers_detail=False,
            ):
        """Return unread plurks on timeline.

        Args:
            offset: return plurks older than offset, formatted as `YYYY-mm-DDTHH:MM:SS`. The function can also
                accept a datetime object and convert it automatically to the format for the API.
            limit: how many plurks should be returned. Default: `20`.
            filter: can be `my`, `responded`, `private`, `favorite`, `replurked` or `mentioned`; `mentioned` is only
                available for users with `user.premium = true`.
            favorers_detail: If true, detailed users information about all favorers of this plurk will be included into
                "plurk_users".
            limited_tail: If true, detailed users information about recepients of this plurk will be included into
                "plurk_users" (if this plurk is private).
            replurkers_detail: If true, detailed users information about all replurkers of this plurk will be included
                into "plurk_users".
        """
        endpoint = f'{self.client.base_url}/APP/Timeline/getUnreadPlurks'
        if isinstance(offset, datetime):
            offset = offset.strftime('%Y-%m-%dT%H:%M:%S')
        params = build_params(
            offset=offset,
            limit=limit,
            filter=filter,
            favorers_detail=favorers_detail,
            limited_detail=limited_detail,
            replurkers_detail=replurkers_detail,
        )
        resp = self.client.http_client.get(endpoint, params=params)
        validate_resp(resp)
        return GetPlurksResp(**resp.json())

    def get_public_plurks(
                self,
                user_id: Union[int, str],
                offset: Optional[Union[str, datetime]] = None,
                limit: int = 20,
                filter: Optional[Filter] = None,  # pylint: disable=redefined-builtin
                favorers_detail=False,
                limited_detail=False,
                replurkers_detail=False,
            ):
        """Return public plurks of a specific user on timeline.

        Args:
            user_id: the user_id of the public plurks owner to get, can be the unique user_id (int) or nick_name (str)
            offset: return plurks older than offset, formatted as `YYYY-mm-DDTHH:MM:SS`. The function can also
                accept a datetime object and convert it automatically to the format for the API.
            limit: how many plurks should be returned. Default: `20`.
            filter: can be `my`, `responded`, `private`, `favorite`, `replurked` or `mentioned`; `mentioned` is only
                available for users with `user.premium = true`.
            favorers_detail: If true, detailed users information about all favorers of this plurk will be included into
                "plurk_users".
            limited_tail: If true, detailed users information about recepients of this plurk will be included into
                "plurk_users" (if this plurk is private).
            replurkers_detail: If true, detailed users information about all replurkers of this plurk will be included
                into "plurk_users".
        """
        endpoint = f'{self.client.base_url}/APP/Timeline/getPublicPlurks'
        if isinstance(offset, datetime):
            offset = offset.strftime('%Y-%m-%dT%H:%M:%S')
        params = build_params(
            user_id=user_id,
            offset=offset,
            limit=limit,
            filter=filter,
            favorers_detail=favorers_detail,
            limited_detail=limited_detail,
            replurkers_detail=replurkers_detail,
        )
        resp = self.client.http_client.get(endpoint, params=params)
        validate_resp(resp)
        return GetPlurksResp(**resp.json())

    def plurk_add(
                self,
                content: str,
                qualifier: Qualifier = Qualifier.EMPTY,
                limited_to: Optional[List[int]] = None,
                no_comments: Optional[int] = None,
                lang: Optional[Language] = None,
            ):
        """Add a new plurk.

        Args:
            content: the content of the plurk.
            qualifier: the qualifier of the plurk.
            limited_to: a list of user_id. if provided, the plurk will be limited to the list of users.
            no_comments: set to 1 to disable all responses, set to 2 to restrict responses to friend-only.
            lang: the plurk's language. The qualifier will be displayed in the specified language.
        """
        endpoint = f'{self.client.base_url}/APP/Timeline/plurkAdd'
        payload = build_params(
            content=content,
            qualifier=qualifier,
            limited_to=str(limited_to),  # API endpoint expects this argument to be stringified
            no_comments=no_comments,
            lang=lang,
        )
        resp = self.client.http_client.post(endpoint, json=payload)
        validate_resp(resp)
        return Plurk(**resp.json())

    def plurk_delete(self, plurk_id: int):
        """Delete an existing plurk.

        WARNING: the endpoint always returns a `success_text='ok'` response, even when providing a plurk_id
        from a plurk the current user can't delete or already deleted.
        To ensure the plurk is successfully deleted, call get_plurk(plurk_id) and catch the BadRequestError.

        Args:
            plurk_id: the unique id of the plurk.
        """
        endpoint = f'{self.client.base_url}/APP/Timeline/plurkDelete'
        payload = build_params(plurk_id=plurk_id)
        resp = self.client.http_client.post(endpoint, json=payload)
        validate_resp(resp)
        return ActionResp(**resp.json())

    def plurk_edit(self, plurk_id: int, content: str, limited_to: Optional[List[int]] = None):
        """"Edit an existing plurk.

        Args:
            plurk_id: the unique id of the plurk.
            content: the new content of the plurk.
        """
        endpoint = f'{self.client.base_url}/APP/Timeline/plurkEdit'
        payload = build_params(plurk_id=plurk_id, content=content, limited_to=str(limited_to))
        resp = self.client.http_client.post(endpoint, json=payload)
        validate_resp(resp)
        return Plurk(**resp.json())

    def toggle_comments(self, plurk_id: int, no_comments: int):
        """Edit a plurk's no_comments value.

        Possbile values:
        - `0`: everyone can comment (default)
        - `1`: no one can comment
        - `2`: only friends can comment
        """
        endpoint = f'{self.client.base_url}/APP/Timeline/toggleComments'
        payload = build_params(plurk_id=plurk_id, no_comments=no_comments)
        resp = self.client.http_client.post(endpoint, json=payload)
        validate_resp(resp)
        return ActionResp(**resp.json())

    def mute_plurks(self, ids: List[int]):
        """Mute plurks.

        Effectively set all plurks' `is_unread` to `2`.
        """
        endpoint = f'{self.client.base_url}/APP/Timeline/mutePlurks'
        payload = build_params(ids=str(ids))
        resp = self.client.http_client.post(endpoint, json=payload)
        validate_resp(resp)
        return ActionResp(**resp.json())

    def unmute_plurks(self, ids: List[int]):
        """Unmute plurks.

        Effectively set all plurks `is_unread` back to `0` or `1`, based on what the original value is.
        Plurk seems to have internally stored the original `is_unread` value before a plurk is muted.
        """
        endpoint = f'{self.client.base_url}/APP/Timeline/unmutePlurks'
        payload = build_params(ids=str(ids))
        resp = self.client.http_client.post(endpoint, json=payload)
        validate_resp(resp)
        return ActionResp(**resp.json())

    def favorite_plurks(self, ids: List[int]):
        """Favorite plurks.

        Effectively set all plurks' `favorite` to `True`.
        """
        endpoint = f'{self.client.base_url}/APP/Timeline/favoritePlurks'
        payload = build_params(ids=str(ids))
        resp = self.client.http_client.post(endpoint, json=payload)
        validate_resp(resp)
        return ActionResp(**resp.json())

    def unfavorite_plurks(self, ids: List[int]):
        """Unfavorite plurks.

        Effectively set all plurks' `favorite` to `True`.
        """
        endpoint = f'{self.client.base_url}/APP/Timeline/unfavoritePlurks'
        payload = build_params(ids=str(ids))
        resp = self.client.http_client.post(endpoint, json=payload)
        validate_resp(resp)
        return ActionResp(**resp.json())

    def replurk(self, ids: List[int]):
        """Replurk plurks.
        """
        endpoint = f'{self.client.base_url}/APP/Timeline/replurk'
        payload = build_params(ids=str(ids))
        resp = self.client.http_client.post(endpoint, json=payload)
        validate_resp(resp)
        return ActionResp(**resp.json())

    def unreplurk(self, ids: List[int]):
        """Unreplurk plurks
        """
        endpoint = f'{self.client.base_url}/APP/Timeline/unreplurk'
        payload = build_params(ids=str(ids))
        resp = self.client.http_client.post(endpoint, json=payload)
        validate_resp(resp)
        return ActionResp(**resp.json())

    def mark_as_read(self, ids: List[int]):
        """Mark plurks as read.

        Effectively set all plurks' `is_unread` to `0`.
        Note that Plurk does not provide any API to undo this action.
        """
        endpoint = f'{self.client.base_url}/APP/Timeline/markAsRead'
        payload = build_params(ids=str(ids))
        resp = self.client.http_client.post(endpoint, json=payload)
        validate_resp(resp)
        return ActionResp(**resp.json())

    def upload_picture(self, image: Union[str, BinaryIO]):
        """Upload an image to Plurk site and get a Plurk CDN image link.

        The provided file must be a valid image, otherwise it will be rejected.
        """
        endpoint = f'{self.client.base_url}/APP/Timeline/uploadPicture'
        if isinstance(image, str):
            with open(image, 'rb') as file_object:
                resp = self.client.http_client.post(
                    endpoint, files={'image': ('image', file_object)}
                )
        else:
            resp = self.client.http_client.post(
                endpoint, files={'image': ('image', image)}
            )
        validate_resp(resp)
        return UploadPictureResp(**resp.json())

    def report_abuse(self, plurk_id: int, category: AbuseCategory):
        """Report a plurk as abuse.
        """
        endpoint = f'{self.client.base_url}/APP/Timeline/reportAbuse'
        payload = build_params(plurk_id=plurk_id, category=category)
        resp = self.client.http_client.post(endpoint, json=payload)
        validate_resp(resp)
        return ActionResp(**resp.json())



from pytest_httpx import HTTPXMock
from pytest_mock import MockerFixture

from plurk import Client
from plurk.enums import AbuseCategory
from plurk.models import (ActionResp, GetPlurkResp, GetPlurksResp, Plurk,
                          UploadPictureResp)

FAKE_APP_CREDENTIALS = ('fake_app_key', 'fake_app_secret')
FAKE_CONTENT = 'foobar'
FAKE_PLURK_ID = -1
FAKE_USER_ID = -1
FAKE_FILEPATH = '/dev/null'


def test_get_plurk(httpx_mock: HTTPXMock, get_plurk_resp_fixture: GetPlurkResp):
    resp_json = get_plurk_resp_fixture.dict_original()
    httpx_mock.add_response(
        status_code=200,
        json=resp_json,
    )
    with Client(*FAKE_APP_CREDENTIALS) as client:
        get_plurk_resp = client.timeline.get_plurk(FAKE_PLURK_ID)
    assert isinstance(get_plurk_resp, GetPlurkResp)
    assert get_plurk_resp.dict_original() == resp_json


def test_get_plurks(httpx_mock: HTTPXMock, get_plurks_resp_fixture: GetPlurksResp):
    resp_json = get_plurks_resp_fixture.dict_original()
    httpx_mock.add_response(
        status_code=200,
        json=resp_json,
    )
    with Client(*FAKE_APP_CREDENTIALS) as client:
        get_plurks_resp = client.timeline.get_plurks()
    assert isinstance(get_plurks_resp, GetPlurksResp)
    assert get_plurks_resp.dict_original() == resp_json


def test_get_unread_plurks(httpx_mock: HTTPXMock, get_plurks_resp_fixture: GetPlurksResp):
    resp_json = get_plurks_resp_fixture.dict_original()
    httpx_mock.add_response(
        status_code=200,
        json=resp_json,
    )
    with Client(*FAKE_APP_CREDENTIALS) as client:
        get_plurks_resp = client.timeline.get_unread_plurks()
    assert isinstance(get_plurks_resp, GetPlurksResp)
    assert get_plurks_resp.dict_original() == resp_json


def test_get_public_plurks(httpx_mock: HTTPXMock, get_plurks_resp_fixture: GetPlurksResp):
    resp_json = get_plurks_resp_fixture.dict_original()
    httpx_mock.add_response(
        status_code=200,
        json=resp_json,
    )
    with Client(*FAKE_APP_CREDENTIALS) as client:
        get_plurks_resp = client.timeline.get_public_plurks(FAKE_USER_ID)
    assert isinstance(get_plurks_resp, GetPlurksResp)
    assert get_plurks_resp.dict_original() == resp_json


def test_plurk_add(httpx_mock: HTTPXMock, plurk_fixture: Plurk):
    resp_json = plurk_fixture.dict_original()
    httpx_mock.add_response(
        status_code=200,
        json=resp_json,
    )
    with Client(*FAKE_APP_CREDENTIALS) as client:
        plurk = client.timeline.plurk_add(FAKE_CONTENT)
    assert isinstance(plurk, Plurk)
    assert plurk.dict_original() == resp_json


def test_plurk_delete(httpx_mock: HTTPXMock, action_resp_fixture: ActionResp):
    resp_json = action_resp_fixture.dict_original()
    httpx_mock.add_response(
        status_code=200,
        json=resp_json,
    )
    with Client(*FAKE_APP_CREDENTIALS) as client:
        action_resp = client.timeline.plurk_delete(FAKE_PLURK_ID)
    assert isinstance(action_resp, ActionResp)
    assert action_resp.dict_original() == resp_json


def test_plurk_edit(httpx_mock: HTTPXMock, plurk_fixture: Plurk):
    resp_json = plurk_fixture.dict_original()
    httpx_mock.add_response(
        status_code=200,
        json=resp_json,
    )
    with Client(*FAKE_APP_CREDENTIALS) as client:
        plurk = client.timeline.plurk_edit(FAKE_PLURK_ID, FAKE_CONTENT)
    assert isinstance(plurk, Plurk)
    assert plurk.dict_original() == resp_json


def test_toggle_comments(httpx_mock: HTTPXMock, action_resp_fixture: ActionResp):
    resp_json = action_resp_fixture.dict_original()
    httpx_mock.add_response(
        status_code=200,
        json=resp_json,
    )
    with Client(*FAKE_APP_CREDENTIALS) as client:
        action_resp = client.timeline.toggle_comments(FAKE_PLURK_ID, 0)
    assert isinstance(action_resp, ActionResp)
    assert action_resp.dict_original() == resp_json


def test_mute_plurks(httpx_mock: HTTPXMock, action_resp_fixture: ActionResp):
    resp_json = action_resp_fixture.dict_original()
    httpx_mock.add_response(
        status_code=200,
        json=resp_json,
    )
    with Client(*FAKE_APP_CREDENTIALS) as client:
        action_resp = client.timeline.mute_plurks([FAKE_PLURK_ID])
    assert isinstance(action_resp, ActionResp)
    assert action_resp.dict_original() == resp_json


def test_unmute_plurks(httpx_mock: HTTPXMock, action_resp_fixture: ActionResp):
    resp_json = action_resp_fixture.dict_original()
    httpx_mock.add_response(
        status_code=200,
        json=resp_json,
    )
    with Client(*FAKE_APP_CREDENTIALS) as client:
        action_resp = client.timeline.unmute_plurks([FAKE_PLURK_ID])
    assert isinstance(action_resp, ActionResp)
    assert action_resp.dict_original() == resp_json


def test_favorite_plurks(httpx_mock: HTTPXMock, action_resp_fixture: ActionResp):
    resp_json = action_resp_fixture.dict_original()
    httpx_mock.add_response(
        status_code=200,
        json=resp_json,
    )
    with Client(*FAKE_APP_CREDENTIALS) as client:
        action_resp = client.timeline.favorite_plurks([FAKE_PLURK_ID])
    assert isinstance(action_resp, ActionResp)
    assert action_resp.dict_original() == resp_json


def test_unfavorite_plurks(httpx_mock: HTTPXMock, action_resp_fixture: ActionResp):
    resp_json = action_resp_fixture.dict_original()
    httpx_mock.add_response(
        status_code=200,
        json=resp_json,
    )
    with Client(*FAKE_APP_CREDENTIALS) as client:
        action_resp = client.timeline.unfavorite_plurks([FAKE_PLURK_ID])
    assert isinstance(action_resp, ActionResp)
    assert action_resp.dict_original() == resp_json


def test_replurk(httpx_mock: HTTPXMock, action_resp_fixture: ActionResp):
    resp_json = action_resp_fixture.dict_original()
    httpx_mock.add_response(
        status_code=200,
        json=resp_json,
    )
    with Client(*FAKE_APP_CREDENTIALS) as client:
        action_resp = client.timeline.replurk([FAKE_PLURK_ID])
    assert isinstance(action_resp, ActionResp)
    assert action_resp.dict_original() == resp_json


def test_unreplurk(httpx_mock: HTTPXMock, action_resp_fixture: ActionResp):
    resp_json = action_resp_fixture.dict_original()
    httpx_mock.add_response(
        status_code=200,
        json=resp_json,
    )
    with Client(*FAKE_APP_CREDENTIALS) as client:
        action_resp = client.timeline.unreplurk([FAKE_PLURK_ID])
    assert isinstance(action_resp, ActionResp)
    assert action_resp.dict_original() == resp_json


def test_mark_as_read(httpx_mock: HTTPXMock, action_resp_fixture: ActionResp):
    resp_json = action_resp_fixture.dict_original()
    httpx_mock.add_response(
        status_code=200,
        json=resp_json,
    )
    with Client(*FAKE_APP_CREDENTIALS) as client:
        action_resp = client.timeline.mark_as_read([FAKE_PLURK_ID])
    assert isinstance(action_resp, ActionResp)
    assert action_resp.dict_original() == resp_json


def test_upload_picture(httpx_mock: HTTPXMock, mocker: MockerFixture, upload_picture_resp_fixture: UploadPictureResp):
    resp_json = upload_picture_resp_fixture.dict_original()
    httpx_mock.add_response(
        status_code=200,
        json=resp_json,
    )
    mocker.patch('builtins.open', mocker.mock_open(read_data=b''))
    with Client(*FAKE_APP_CREDENTIALS) as client:
        upload_picture_resp = client.timeline.upload_picture(FAKE_FILEPATH)
    assert isinstance(upload_picture_resp, UploadPictureResp)
    assert upload_picture_resp.dict_original() == resp_json


def test_report_abuse(httpx_mock: HTTPXMock, action_resp_fixture: ActionResp):
    resp_json = action_resp_fixture.dict_original()
    httpx_mock.add_response(
        status_code=200,
        json=resp_json,
    )
    with Client(*FAKE_APP_CREDENTIALS) as client:
        action_resp = client.timeline.report_abuse(FAKE_PLURK_ID, AbuseCategory.SPAM)
    assert isinstance(action_resp, ActionResp)
    assert action_resp.dict_original() == resp_json



from typing import Dict, List

from plurk.models.base import RespBase
from plurk.models.plurk import Plurk
from plurk.models.user_data import SimpleUserData


class GetPlurkResp(RespBase):
    """Resp of `/APP/Timeline/getPlurk`
    """
    plurk: Plurk
    user: SimpleUserData
    plurk_users: Dict[str, SimpleUserData]


class GetPlurksResp(RespBase):
    """Resp of
    - `/APP/Timeline/getPlurks`
    - `/APP/Timeline/getUnreadPlurks`
    - `/APP/Timeline/getPublicPlurks`
    """
    plurks: List[Plurk]
    plurk_users: Dict[str, SimpleUserData]


class UploadPictureResp(RespBase):
    """Resp of `/APP/Timeline/uploadPicture`
    """
    full: str
    thumbnail: str



from .alerts import Alerts
from .blocks import Blocks
from .cliques import Cliques
from .emoticons import Emoticons
from .friends_fans import FriendsFans
from .helper import Helper
from .plurk_search import PlurkSearch
from .profile import Profile
from .realtime import Realtime
from .responses import Responses
from .timeline import Timeline
from .user_search import UserSearch
from .users import Users

__all__ = [
    'Alerts',
    'Blocks',
    'Cliques',
    'Emoticons',
    'Helper',
    'FriendsFans',
    'PlurkSearch',
    'Profile',
    'Realtime',
    'Responses',
    'Timeline',
    'Users',
    'UserSearch',
]



from datetime import datetime
from typing import Literal

from plurk.enums import AlertType
from plurk.models.base import RespBase
from plurk.models.user_data import PublicUserData
from plurk.utils import parse_time_validator


class AlertBase(RespBase):
    """
    General data structures
    The data returned by getActive and getHistory can be of following nature:
    Friendship request: (requires action from the user)
    {"type": "friendship_request", "from_user": {"nick_name": ...}, "posted": ...}

    Friendship pending: (requires action from the user)
    {"type": "friendship_pending", "to_user": {"nick_name": ...}, "posted": ...}

    """
    type: AlertType
    posted: datetime

    _parse_posted = parse_time_validator('posted')


class FriendshipRequestAlert(AlertBase):
    type: Literal[AlertType.FRIENDSHIP_REQUEST]
    from_user: PublicUserData


class FriendshipPendingAlert(AlertBase):
    type: Literal[AlertType.FRIENDSHIP_PENDING]
    to_user: PublicUserData


class NewFanAlert(AlertBase):
    type: Literal[AlertType.NEW_FAN]
    new_fan: PublicUserData


class FriendshipAcceptedAlert(AlertBase):
    type: Literal[AlertType.FRIENDSHIP_ACCEPTED]
    friend_info: PublicUserData


class NewFriendAlert(AlertBase):
    type: Literal[AlertType.NEW_FRIEND]
    new_friend: PublicUserData


class PrivatePlurkAlert(AlertBase):
    type: Literal[AlertType.PRIVATE_PLURK]
    owner: PublicUserData
    plurk_id: int


class PlurkLikedAlert(AlertBase):
    type: Literal[AlertType.PLURK_LIKED]
    from_user: PublicUserData
    plurk_id: int
    num_others: int


class PlurkReplurkedAlert(AlertBase):
    type: Literal[AlertType.PLURK_REPLURKED]
    from_user: PublicUserData
    plurk_id: int
    num_others: int


class MentionedAlert(AlertBase):
    type: Literal[AlertType.MENTIONED]
    from_user: PublicUserData
    plurk_id: int
    num_others: int
    response_id: int


class MyRespondedAlert(AlertBase):
    type: Literal[AlertType.MY_RESPONDED]
    from_user: PublicUserData
    plurk_id: int
    num_others: int
    response_id: int



from pytest_httpx import HTTPXMock
from pytest_mock import MockerFixture

from plurk import Client
from plurk.models import KarmaStats, UpdateAvatarResp, UserData


FAKE_APP_CREDENTIALS = ('fake_app_key', 'fake_app_secret')
FAKE_FILEPATH = '/dev/null'


def test_me(httpx_mock: HTTPXMock, user_data_fixture):
    resp_json = user_data_fixture.dict_original()
    httpx_mock.add_response(
        status_code=200,
        json=resp_json,
    )
    with Client(*FAKE_APP_CREDENTIALS) as client:
        user_data = client.users.me()
    assert isinstance(user_data, UserData)
    assert user_data.dict_original() == resp_json


def test_update_avatar(httpx_mock: HTTPXMock, mocker: MockerFixture, update_avatar_resp_fixture):
    resp_json = update_avatar_resp_fixture.dict_original()
    httpx_mock.add_response(
        status_code=200,
        json=resp_json,
    )
    mocker.patch('builtins.open', mocker.mock_open(read_data=b''))
    with Client(*FAKE_APP_CREDENTIALS) as client:
        update_avatar_resp = client.users.update_avatar(FAKE_FILEPATH)
    assert isinstance(update_avatar_resp, UpdateAvatarResp)
    assert update_avatar_resp.dict_original() == resp_json


def test_get_karma_stats(httpx_mock: HTTPXMock, karma_stats_fixture):
    resp_json = karma_stats_fixture.dict_original()
    httpx_mock.add_response(
        status_code=200,
        json=resp_json,
    )
    with Client(*FAKE_APP_CREDENTIALS) as client:
        karma_stats = client.users.get_karma_stats()
    assert isinstance(karma_stats, KarmaStats)
    assert karma_stats.dict_original() == resp_json



from typing import Dict, List

from plurk.models.base import RespBase
from plurk.models.plurk import Plurk
from plurk.models.user_data import SimpleUserData


class PlurkSearchResp(RespBase):
    """Resp of `/APP/PlurkSearch/search`
    """
    plurks: List[Plurk]
    users: Dict[str, SimpleUserData]


class UserSearchResp(RespBase):
    """Resp of `/APP/UserSearch/search`
    """
    counts: int
    users: List[SimpleUserData]



from typing import Dict

from plurk.apis.base import BaseApi
from plurk.exceptions import validate_resp


class Emoticons(BaseApi):
    def get(self) -> Dict[str, Dict]:
        r"""Emoticons are a big part of Plurk since they make it easy to express feelings.

        This call returns a JSON object that looks like:
        ```
        {"karma": {"0": [[":-))", "https:\/\/s.plurk.com\/xxxxx.gif"], ...], ...},
        "recruited": {"10": [["(bigeyes)", "https:\/\/s.plurk.com\/xxxxx.gif"], ...], ...} }
        ```
        """
        endpoint = f'{self.client.base_url}/APP/Emoticons/get'
        resp = self.client.http_client.get(endpoint)
        validate_resp(resp)
        return resp.json()



from typing import Dict, Optional

from plurk.models.base import RespBase


class ActionResp(RespBase):
    """Resp of endpoints that performs an action and only return the success status:
    - `/APP/Timeline/plurkDelete`
    - `/APP/Timeline/toggleComments`
    - `/APP/Timeline/mutePlurks`
    - `/APP/Timeline/unmutePlurks`
    - `/APP/Timeline/favoritePlurks`
    - `/APP/Timeline/unfavoritePlurks`
    - `/APP/Timeline/replurk`
    - `/APP/Timeline/unreplurk`
    - `/APP/Timeline/markAsRead`
    - `/APP/Timeline/reportAbuse`
    - `/APP/Responses/responseDelete`
    """
    success_text: Optional[str]
    success: Optional[bool]
    results: Optional[Dict[str, Dict]]

    def is_successful(self):
        return self.success_text or self.success is not False



from typing import Dict, List

from pytest_httpx import HTTPXMock

from plurk import Client
from plurk.models import ActionResp, PublicUserData

FAKE_APP_CREDENTIALS = ('fake_app_key', 'fake_app_secret')
FAKE_CONTENT = 'foobar'
FAKE_USER_ID = -1


def test_get_friends_by_offset(httpx_mock: HTTPXMock, public_user_data_list_fixture: List[PublicUserData]):
    resp_json = [i.dict_original() for i in public_user_data_list_fixture]
    httpx_mock.add_response(
        status_code=200,
        json=resp_json,
    )
    with Client(*FAKE_APP_CREDENTIALS) as client:
        resp = client.friends_fans.get_friends_by_offset(FAKE_USER_ID)
    assert isinstance(resp, List)
    assert [i.dict_original() for i in resp] == resp_json


def test_get_fans_by_offset(httpx_mock: HTTPXMock, public_user_data_list_fixture: List[PublicUserData]):
    resp_json = [i.dict_original() for i in public_user_data_list_fixture]
    httpx_mock.add_response(
        status_code=200,
        json=resp_json,
    )
    with Client(*FAKE_APP_CREDENTIALS) as client:
        resp = client.friends_fans.get_fans_by_offset(FAKE_USER_ID)
    assert isinstance(resp, List)
    assert [i.dict_original() for i in resp] == resp_json
    

def test_get_following_by_offset(httpx_mock: HTTPXMock, public_user_data_list_fixture: List[PublicUserData]):
    resp_json = [i.dict_original() for i in public_user_data_list_fixture]
    httpx_mock.add_response(
        status_code=200,
        json=resp_json,
    )
    with Client(*FAKE_APP_CREDENTIALS) as client:
        resp = client.friends_fans.get_following_by_offset(FAKE_USER_ID)
    assert isinstance(resp, List)
    assert [i.dict_original() for i in resp] == resp_json


def test_become_friend(httpx_mock: HTTPXMock, action_resp_fixture: ActionResp):
    resp_json = action_resp_fixture.dict_original()
    httpx_mock.add_response(
        status_code=200,
        json=resp_json,
    )
    with Client(*FAKE_APP_CREDENTIALS) as client:
        resp = client.friends_fans.become_friend(FAKE_USER_ID)
    assert isinstance(resp, ActionResp)
    assert resp.dict_original() == resp_json


def test_remove_as_friend(httpx_mock: HTTPXMock, action_resp_fixture: ActionResp):
    resp_json = action_resp_fixture.dict_original()
    httpx_mock.add_response(
        status_code=200,
        json=resp_json,
    )
    with Client(*FAKE_APP_CREDENTIALS) as client:
        resp = client.friends_fans.remove_as_friend(FAKE_USER_ID)
    assert isinstance(resp, ActionResp)
    assert resp.dict_original() == resp_json


def test_become_fan(httpx_mock: HTTPXMock, action_resp_fixture: ActionResp):
    resp_json = action_resp_fixture.dict_original()
    httpx_mock.add_response(
        status_code=200,
        json=resp_json,
    )
    with Client(*FAKE_APP_CREDENTIALS) as client:
        resp = client.friends_fans.become_fan(FAKE_USER_ID)
    assert isinstance(resp, ActionResp)
    assert resp.dict_original() == resp_json


def test_set_following(httpx_mock: HTTPXMock, action_resp_fixture: ActionResp):
    resp_json = action_resp_fixture.dict_original()
    httpx_mock.add_response(
        status_code=200,
        json=resp_json,
    )
    with Client(*FAKE_APP_CREDENTIALS) as client:
        resp = client.friends_fans.set_following(FAKE_USER_ID, False)
    assert isinstance(resp, ActionResp)
    assert resp.dict_original() == resp_json


def test_remove_as_fan(httpx_mock: HTTPXMock, action_resp_fixture: ActionResp):
    resp_json = action_resp_fixture.dict_original()
    httpx_mock.add_response(
        status_code=200,
        json=resp_json,
    )
    with Client(*FAKE_APP_CREDENTIALS) as client:
        resp = client.friends_fans.remove_as_fan(FAKE_USER_ID)
    assert isinstance(resp, ActionResp)
    assert resp.dict_original() == resp_json


def test_get_completion(httpx_mock: HTTPXMock):
    resp_json = {
        '123': {
            'full_name': 'Plurk User',
            'nick_name': 'plurk',
            'avatar': 12345,
            'display_name': 'Plurker'
        }
    }
    httpx_mock.add_response(
        status_code=200,
        json=resp_json,
    )
    with Client(*FAKE_APP_CREDENTIALS) as client:
        resp = client.friends_fans.get_completion()
    assert isinstance(resp, Dict)
    assert resp == resp_json



from pytest_httpx import HTTPXMock

from plurk import Client
from plurk.models import PlurkSearchResp

FAKE_APP_CREDENTIALS = ('fake_app_key', 'fake_app_secret')


def test_search(httpx_mock: HTTPXMock, plurk_search_resp_fixture):
    resp_json = plurk_search_resp_fixture.dict_original()
    httpx_mock.add_response(
        status_code=200,
        json=resp_json,
    )
    with Client(*FAKE_APP_CREDENTIALS) as client:
        plurk_search_resp = client.plurk_search.search('')
    assert isinstance(plurk_search_resp, PlurkSearchResp)
    assert plurk_search_resp_fixture.dict_original() == resp_json



from pytest_httpx import HTTPXMock

from plurk import Client
from plurk.models import UserChannel


FAKE_APP_CREDENTIALS = ('fake_app_key', 'fake_app_secret')


def test_get_user_channel(httpx_mock: HTTPXMock, user_channel_fixture):
    resp_json = user_channel_fixture.dict_original()
    httpx_mock.add_response(
        status_code=200,
        json=resp_json,
    )
    with Client(*FAKE_APP_CREDENTIALS) as client:
        user_channel = client.realtime.get_user_channel()
    assert isinstance(user_channel, UserChannel)
    assert user_channel.dict_original() == resp_json



from datetime import datetime
from typing import Any, List, Literal, Optional

from plurk.enums import ChannelDataEntryType
from plurk.models.base import RespBase
from plurk.utils import parse_time_validator


class Plurk(RespBase):
    """Data of a plurk.

    A plurk's data returned from different endpoint varies.
    The official doc does not describe the exact return value of each endpoint. If an optional
    field is absent in the data, it might be available through other endpoints or it might simply not available.
    """
    plurk_id: int
    content: str
    content_raw: str
    qualifier: str
    posted: datetime
    last_edited: Optional[datetime]
    owner_id: Optional[int]
    user_id: int
    response_count: int
    limited_to: Optional[Any]  # ?
    excluded: Optional[Any]  # ?
    lang: str
    is_unread: int
    no_comments: int
    replurked: bool
    replurker_id: Optional[int]
    replurkers: List[int]
    replurkers_count: int
    favorers: List[int]
    favorite: bool
    favorite_count: int
    anonymous: bool
    plurk_type: int
    coins: int
    id: Optional[int]
    responses_seen: int
    porn: bool
    publish_to_followers: bool
    with_poll: bool
    has_gift: bool
    responded: int
    mentioned: int
    bookmark: Optional[bool]
    qualifier_translated: Optional[str]
    replurkable: Optional[bool]

    _parse_last_edited = parse_time_validator('last_edited')
    _parse_posted = parse_time_validator('posted')


class NewPlurk(Plurk):
    """Plurk data returned from comet channel URL
    """
    type: Literal[ChannelDataEntryType.NEW_PLURK]



from pytest_httpx import HTTPXMock

from plurk import Client
from plurk.models import ActionResp, Response, ResponsesGetResp

FAKE_APP_CREDENTIALS = ('fake_app_key', 'fake_app_secret')
FAKE_CONTENT = 'foobar'
FAKE_PLURK_ID = -1
FAKE_RESPONSE_ID = -1


def test_get(httpx_mock: HTTPXMock, responses_get_resp_fixture: ResponsesGetResp):
    resp_json = responses_get_resp_fixture.dict_original()
    httpx_mock.add_response(
        status_code=200,
        json=resp_json,
    )
    with Client(*FAKE_APP_CREDENTIALS) as client:
        resp = client.responses.get(FAKE_PLURK_ID)
    assert isinstance(resp, ResponsesGetResp)
    assert resp.dict_original() == resp_json


def test_response_add(httpx_mock: HTTPXMock, response_fixture: Response):
    resp_json = response_fixture.dict_original()
    httpx_mock.add_response(
        status_code=200,
        json=resp_json,
    )
    with Client(*FAKE_APP_CREDENTIALS) as client:
        response = client.responses.response_add(FAKE_PLURK_ID, FAKE_CONTENT)
    assert isinstance(response, Response)
    assert response.dict_original() == resp_json


def test_response_delete(httpx_mock: HTTPXMock, action_resp_fixture: ActionResp):
    resp_json = action_resp_fixture.dict_original()
    httpx_mock.add_response(
        status_code=200,
        json=resp_json,
    )
    with Client(*FAKE_APP_CREDENTIALS) as client:
        action_resp = client.responses.response_delete(FAKE_PLURK_ID, FAKE_RESPONSE_ID)
    assert isinstance(action_resp, ActionResp)
    assert action_resp.dict_original() == resp_json



from typing import Dict

from authlib.integrations.httpx_client import OAuth1Client


def get_oauth_client(
    client_id: str, client_secret: str,
    token=None, token_secret=None,
    redirect_uri=None,
):
    return OAuth1Client(
        client_id, client_secret,
        token=token, token_secret=token_secret,
        redirect_uri=redirect_uri,
    )


def get_request_token(client: OAuth1Client, request_token_url: str):
    return client.fetch_request_token(request_token_url)


def get_auth_url(client: OAuth1Client, authenticate_url: str, request_token: Dict[str, str]):
    return client.create_authorization_url(authenticate_url, request_token['oauth_token'])


def fetch_access_token(
            client_id: str, client_secret: str, access_token_url: str,
            request_token: Dict[str, str], oauth_verifier: str,
        ):
    with OAuth1Client(
        client_id,
        client_secret,
        token=request_token['oauth_token'],
        token_secret=request_token['oauth_token_secret']
    ) as client:
        token = client.fetch_access_token(access_token_url, oauth_verifier)
    return token



from typing import Dict, List

from plurk.enums import Privacy
from plurk.models.base import RespBase
from plurk.models.plurk import Plurk
from plurk.models.user_data import PublicUserData, UserData


class ProfileBase(RespBase):
    """Base class for profile
    """
    fans_count: int
    friends_count: int
    privacy: Privacy
    plurks: List[Plurk]
    user_info: UserData


class OwnProfile(ProfileBase):
    """Data returned by `/APP/Profile/getOwnProfile` endpoint
    """
    alerts_count: int
    has_read_permission: bool
    plurks_users: Dict[str, PublicUserData]
    """A dict of public user data.

    Each key is a stringified user id and the value is its public user data.
    """
    unread_count: int


class PublicProfile(ProfileBase):
    """Data returned by `/APP/Profile/getPublicProfile` endpoint

    WARNING: The data model has not yet been verified.
    """



from enum import Enum


class AlertType(str, Enum):
    FRIENDSHIP_REQUEST = 'friendship_request'
    FRIENDSHIP_PENDING = 'friendship_pending'
    NEW_FAN = 'new_fan'
    FRIENDSHIP_ACCEPTED = 'friendship_accepted'
    NEW_FRIEND = 'new_friend'
    PRIVATE_PLURK = 'private_plurk'
    PLURK_LIKED = 'plurk_liked'
    PLURK_REPLURKED = 'plurk_replurked'
    MENTIONED = 'mentioned'
    MY_RESPONDED = 'my_responded'



from datetime import datetime
from typing import BinaryIO, Optional, Union

from plurk.apis.base import BaseApi
from plurk.exceptions import validate_resp
from plurk.models.karma_stats import KarmaStats
from plurk.models.user_data import Privacy, UpdateAvatarResp, UserData


class Users(BaseApi):
    def me(self):
        """Returns information about current user.
        """
        endpoint = f'{self.client.base_url}/APP/Users/me'
        resp = self.client.http_client.get(endpoint)
        validate_resp(resp)
        return UserData(**resp.json())

    def update(
        self,
        full_name: Optional[str] = None,
        email: Optional[str] = None,
        display_name: Optional[str] = None,
        privacy: Optional[Privacy] = None,
        date_of_birth: Optional[Union[str, datetime]] = None,
    ):
        """TODO: Question sent to official on 2022/12/17
        wait for official response to solve error
        'only verified third party app can update user info'
        """
        endpoint = f'{self.client.base_url}/APP/Users/update'
        payload = {
            'full_name': full_name,
            'email': email,
            'display_name': display_name,
            'privacy': privacy,
            'date_of_birth': date_of_birth,
        }
        resp = self.client.http_client.post(endpoint, json=payload)
        validate_resp(resp)
        return UserData(**resp.json())

    def update_avatar(self, profile_image: Union[str, BinaryIO]):
        """Update a user's profile picture.

        The picture will be scaled down to 3 versions:
        big, medium and small. The optimal size of `profile_image`
        should be 195x195 pixels.
        """
        endpoint = f'{self.client.base_url}/APP/Users/updateAvatar'
        if isinstance(profile_image, str):
            with open(profile_image, 'rb') as file_object:
                resp = self.client.http_client.post(
                    endpoint, files={'profile_image': ('profile_image', file_object)}
                )
        else:
            resp = self.client.http_client.post(
                endpoint, files={'profile_image': ('profile_image', profile_image)}
            )
        # filename must be included in form-data, otherwise the request will fail.
        # arbitrary filename works, no need to have correct file ext.

        validate_resp(resp)
        return UpdateAvatarResp(**resp.json())

    def get_karma_stats(self):
        """Returns info about current user's karma, including current karma, karma growth,
        karma graph and the latest reason why the karma has dropped.
        """
        endpoint = f'{self.client.base_url}/APP/Users/getKarmaStats'
        resp = self.client.http_client.get(endpoint)
        validate_resp(resp)
        return KarmaStats(**resp.json())



from plurk.apis.base import BaseApi
from plurk.models import ChannelResp


class Helper(BaseApi):
    """The class represents the collection of helper methods build on
    existing Plurk APIs. It does not represent an API endpoint directly.
    """
    def subscribe_to_user_channel(self):
        """Returns a generator that yields plurks posted to the user channel.

        The method uses the `/APP/Realtime/getUserChannel` endpoint to get the current users channel
        and yield received results.

        The generator will yield results indefinitely in a `while True:` loop.
        """
        user_channel = self.client.realtime.get_user_channel()
        new_offset = None
        while True:
            channel_resp = self.client.realtime.get_channel_response(
                user_channel.comet_server, timeout=120, new_offset=new_offset
            )
            new_offset = channel_resp.new_offset
            if isinstance(channel_resp, ChannelResp):
                yield channel_resp



from typing import Any, Dict, List, Optional

from pydantic import parse_obj_as

from plurk.apis.base import BaseApi
from plurk.exceptions import validate_resp
from plurk.models import ActionResp, PublicUserData
from plurk.utils import build_params


class FriendsFans(BaseApi):
    def get_friends_by_offset(self, user_id: int, offset: Optional[int] = None, limit: int = 10):
        """Returns `user_id`'s friend list.

        User can toggle their friend list visibility in privacy settings. If the current user
        can't see the user's friend list, the result will be an empty list.
        """
        endpoint = f'{self.client.base_url}/APP/FriendsFans/getFriendsByOffset'
        params = build_params(user_id=user_id, offset=offset, limit=limit)
        resp = self.client.http_client.get(endpoint, params=params)
        validate_resp(resp)
        return parse_obj_as(List[PublicUserData], resp.json())

    def get_fans_by_offset(self, user_id: int, offset: Optional[int] = None, limit: int = 10):
        """Returns `user_id`'s fan list.

        User can toggle their fan list visibility in privacy settings. If the current user
        can't see the user's fan list, the result will be an empty list.
        """
        endpoint = f'{self.client.base_url}/APP/FriendsFans/getFansByOffset'
        params = build_params(user_id=user_id, offset=offset, limit=limit)
        resp = self.client.http_client.get(endpoint, params=params)
        validate_resp(resp)
        return parse_obj_as(List[PublicUserData], resp.json())

    def get_following_by_offset(self, offset: Optional[int] = None, limit: int = 10):
        """Returns current user's following list.
        """
        endpoint = f'{self.client.base_url}/APP/FriendsFans/getFollowingByOffset'
        params = build_params(offset=offset, limit=limit)
        resp = self.client.http_client.get(endpoint, params=params)
        validate_resp(resp)
        return parse_obj_as(List[PublicUserData], resp.json())

    def become_friend(self, friend_id: int):
        """Create a friend request to `friend_id`. User with `friend_id` has to accept a friendship.
        """
        endpoint = f'{self.client.base_url}/APP/FriendsFans/becomeFriend'
        payload = build_params(friend_id=friend_id)
        resp = self.client.http_client.post(endpoint, json=payload)
        validate_resp(resp)
        return ActionResp(**resp.json())

    def remove_as_friend(self, friend_id: int):
        """Remove friend with ID `friend_id`. `friend_id` won't be notified.
        """
        endpoint = f'{self.client.base_url}/APP/FriendsFans/removeAsFriend'
        payload = build_params(friend_id=friend_id)
        resp = self.client.http_client.post(endpoint, json=payload)
        validate_resp(resp)
        return ActionResp(**resp.json())

    def become_fan(self, fan_id: int):
        """Become fan of fan_id.
        """
        endpoint = f'{self.client.base_url}/APP/FriendsFans/becomeFan'
        payload = build_params(fan_id=fan_id)
        resp = self.client.http_client.post(endpoint, json=payload)
        validate_resp(resp)
        return ActionResp(**resp.json())

    def set_following(self, user_id: int, follow: bool):
        """Update following of user_id. A user can befriend someone, but can unfollow them.
        This request is also used to stop following someone as a fan.
        """
        endpoint = f'{self.client.base_url}/APP/FriendsFans/setFollowing'
        payload = build_params(user_id=user_id, follow=follow)
        resp = self.client.http_client.post(endpoint, json=payload)
        validate_resp(resp)
        return ActionResp(**resp.json())

    def remove_as_fan(self, user_id: int):
        """Stop being a fan of the target user.

        The function is shortcut using the setFollowing endpoint, not mapping to a standalone endpoint.
        """
        return self.set_following(user_id, follow=False)

    def get_completion(self) -> Dict[str, Dict[str, Any]]:
        """Returns a JSON object of the logged in users friends (nick name and full name).
        This information can be used to construct auto-completion for private plurking.
        Notice that a friend list can be big, depending on how many friends a user has,
        so this list should be lazy-loaded in your application.

        Example returning data:
        {
            '123': {
                'full_name': 'Plurk User',
                'nick_name': 'plurk',
                'avatar': 12345,
                'display_name': 'Plurker'
            }
        }

        where '123' is the user id.
        """
        endpoint = f'{self.client.base_url}/APP/FriendsFans/getCompletion'
        resp = self.client.http_client.get(endpoint)
        validate_resp(resp)
        return resp.json()



from datetime import datetime
from typing import Any, List, Optional

from pydantic import BaseModel

from plurk.enums import AvatarSize, Privacy, Relationship
from plurk.models.base import RespBase
from plurk.utils import parse_time_validator


class FilterSetting(BaseModel):
    porn: int
    anonymous: int
    keywords: Optional[str]


class Anniversary(BaseModel):
    years: Optional[int]
    days: Optional[int]


class Birthday(BaseModel):
    day: Optional[int]
    month: Optional[int]
    year: Optional[int]


class BaseUserData(RespBase):
    """Base class for user data
    """
    id: int
    """A unique user id.
    """
    status: str
    has_profile_image: int
    timeline_privacy: int
    nick_name: str
    display_name: Optional[str]
    date_of_birth: Optional[datetime]
    avatar: Optional[int]
    gender: int
    """Gender of the user.

    `0`: female
    `1`: male
    `2`: not stating or other
    """
    karma: float
    premium: bool
    verified_account: bool
    dateformat: int
    default_lang: str
    friend_list_privacy: str
    name_color: Optional[str]
    _version: Optional[str]
    birthday: Optional[Birthday]

    _parse_date_of_birth = parse_time_validator('date_of_birth')

    @property
    def avatar_url_small(self):
        """Constructed avatar URL. It is recommended to use attribute avatar_small instead.
        """
        return self._get_avatar_url(AvatarSize.SMALL)

    @property
    def avatar_url_medium(self):
        """Constructed avatar URL. It is recommended to use attribute avatar_medium instead.
        """
        return self._get_avatar_url(AvatarSize.MEDIUM)

    @property
    def avatar_url_big(self):
        """Constructed avatar URL. It is recommended to use attribute avatar_big instead.
        """
        return self._get_avatar_url(AvatarSize.BIG)

    def _get_avatar_url(self, size: AvatarSize = AvatarSize.BIG):
        """Construct avatar URL based on UserData and desired image size.

        The rule for getting the URL is described in official API doc: https://www.plurk.com/API
        The actual data returned as for now already includes avatar URLs, so there's no need to
        construct URLs now.
        """
        if not self.avatar:
            raise ValueError(f'user {self.id} does not have an avatar')

        url_template = '{base_url}/{user_id}-{size}{avatar}.{ext}'

        if self.has_profile_image == 1:
            base_url = 'https://avatars.plurk.com'
            user_id = self.id
            if size == AvatarSize.BIG:
                ext = 'jpg'
            else:
                ext = 'gif'
        else:
            base_url = 'https://www.plurk.com/static'
            user_id = 'default'
            ext = 'jpg'

        avatar = self.avatar if (self.has_profile_image == 1 and self.avatar) else ''

        return url_template.format(
            base_url=base_url,
            user_id=user_id,
            size=size.value,
            avatar=avatar,
            ext=ext,
        )


class PublicUserData(BaseUserData):
    full_name: str
    location: str
    timezone: Optional[str]
    phone_verified: Optional[int]
    """Phone verification status.

    `None`: phone number is not set.
    `0`: phone number is set but not verified
    `1`: phone verified.
    """
    # not sure; only saw None in real data
    bday_privacy: int
    """The privacy setting for displaying birthday.

    `0`: hide birthday
    `1`: show birth date but not birth year
    `2`: show all
    """
    pinned_plurk_id: Optional[int]
    background_id: int
    show_ads: bool


class UserData(PublicUserData):
    """Data returned by `/APP/Users/me` endpoint
    """
    plurks_count: int
    response_count: int
    profile_views: int
    avatar_small: str
    avatar_medium: str
    avatar_big: str
    about: str
    about_renderred: str
    page_title: str
    recruited: int
    relationship: Relationship
    friends_count: int
    fans_count: int
    join_date: datetime
    hide_plurks_before: Optional[Any]  # not sure; only saw None in real data
    privacy: Privacy
    accept_private_plurk_from: str
    post_anonymous_plurk: bool
    badges: List[str]
    setup_twitter_sync: bool
    filter: FilterSetting
    anniversary: Anniversary
    phone_number: Optional[str]
    creature: int
    creature_url: Optional[str]
    creature_special: int
    creature_special_url: Optional[str]

    _parse_join_date = parse_time_validator('join_date')

    class Config:
        use_enum_values = True


class UpdateAvatarResp(PublicUserData):
    """Data returned by `/APP/Users/updateAvatar` endpoint
    """


class SimpleUserData(BaseUserData):
    """User data in NewPlurk or NewReponse retrieved in user channel data
    """



from typing import List, Union

from pydantic import Field, parse_obj_as
from typing_extensions import Annotated

from plurk.apis.base import BaseApi
from plurk.exceptions import validate_resp
from plurk.models import (ActionResp, FriendshipAcceptedAlert,
                          FriendshipPendingAlert, FriendshipRequestAlert,
                          MentionedAlert, MyRespondedAlert, NewFanAlert,
                          NewFriendAlert, PlurkLikedAlert, PlurkReplurkedAlert,
                          PrivatePlurkAlert)
from plurk.utils import build_params

Alert = Annotated[
    Union[
        FriendshipAcceptedAlert, FriendshipPendingAlert, FriendshipRequestAlert,
        MentionedAlert, MyRespondedAlert, NewFanAlert, NewFriendAlert, PlurkLikedAlert,
        PlurkReplurkedAlert, PrivatePlurkAlert
    ],
    Field(discriminator='type')
]


class Alerts(BaseApi):
    def get_active(self):
        """Return a JSON list of current active alerts.
        """
        endpoint = f'{self.client.base_url}/APP/Alerts/getActive'
        resp = self.client.http_client.get(endpoint)
        validate_resp(resp)
        return parse_obj_as(List[Alert], resp.json())

    def get_history(self):
        """Return a JSON list of past 30 alerts.
        """
        endpoint = f'{self.client.base_url}/APP/Alerts/getHistory'
        resp = self.client.http_client.get(endpoint)
        validate_resp(resp)
        return parse_obj_as(List[Alert], resp.json())

    def add_as_fan(self, user_id: int):
        """Accept user_id as fan.
        """
        endpoint = f'{self.client.base_url}/APP/Alerts/addAsFan'
        payload = build_params(user_id=user_id)
        resp = self.client.http_client.post(endpoint, json=payload)
        validate_resp(resp)
        return ActionResp(**resp.json())

    def add_all_as_fan(self):
        """Accept all friendship requests as fans.
        """
        endpoint = f'{self.client.base_url}/APP/Alerts/addAllAsFan'
        resp = self.client.http_client.post(endpoint)
        validate_resp(resp)
        return ActionResp(**resp.json())

    def add_as_friend(self, user_id: int):
        """Accept user_id as friend.
        """
        endpoint = f'{self.client.base_url}/APP/Alerts/addAsFriend'
        payload = build_params(user_id=user_id)
        resp = self.client.http_client.post(endpoint, json=payload)
        validate_resp(resp)
        return ActionResp(**resp.json())

    def add_all_as_friends(self):
        """Accept all friendship requests as friends.
        """
        endpoint = f'{self.client.base_url}/APP/Alerts/addAllAsFriends'
        resp = self.client.http_client.post(endpoint)
        validate_resp(resp)
        return ActionResp(**resp.json())

    def deny_friendship(self, user_id: int):
        """Accept user_id as friend.
        """
        endpoint = f'{self.client.base_url}/APP/Alerts/denyFriendship'
        payload = build_params(user_id=user_id)
        resp = self.client.http_client.post(endpoint, json=payload)
        validate_resp(resp)
        return ActionResp(**resp.json())

    def remove_notification(self, user_id: int):
        """Accept user_id as friend.
        """
        endpoint = f'{self.client.base_url}/APP/Alerts/removeNotification'
        payload = build_params(user_id=user_id)
        resp = self.client.http_client.post(endpoint, json=payload)
        validate_resp(resp)
        return ActionResp(**resp.json())
